/* %W% %E% */
/*
 * Creation 1986 by Inge B. Kristiansen.
 *
 * 12/04/90 raw: add arguments in externalize/internalize creation.
 * 03.07.90 kk  interrupt interface is the same as procedure -
 *              kept the user interface for backward compatibility.
 * 28.01.93 kk  added ifref() to check if it is a pointer
 */

#include "sysconfig.h"
#include <ctype.h>
#include "np.h"


/*
 * C passes arrays by reference but all other types by value.
 * Check if this is really a pointer to give warning on return params
 */
int
notref(type, address_mode)
	struct object *type;
	enum addr address_mode;
{
	if (type->o_class != O_TYPE)
		error(FATAL, "internal error (refamode): not a type");
	switch( address_mode ) {
		case A_VALUE:
			return (type->t_constr == C_ARRAY ? 0 : 1);

		case A_REF:
			return (0);
	}
}

/*
 * This function is used to cope with the fact that C passes arrays
 * by reference but all other types by value.
 * The argument should be a base type.
 */
char *
refstr(type)
	struct object *type;
{
	if (type->o_class != O_TYPE)
		error(FATAL, "internal error (refstr): not a type");
	return (type->t_constr == C_ARRAY ? "" : "&");
}

/*
 * This function is used to cope with the fact that there might be
 * a forced by reference parameter passing.
 * The argument should be a base type.
 */
char *
refamode(type, address_mode)
	struct object *type;
	enum addr address_mode;
{
	if (type->o_class != O_TYPE)
		error(FATAL, "internal error (refamode): not a type");
	switch( address_mode ) {
		case A_VALUE:
			return (type->t_constr == C_ARRAY ? "" : "&");

		case A_REF:
			return ("");
	}
}

/*
 * This function is used to permute the addressing modes.
 */
enum addr
chamode(address_mode)
	enum addr address_mode;
{
	switch (address_mode) {
		case A_VALUE:
			return (A_REF);

		case A_REF:
			return (A_VALUE);
	}
}

/*
 * This function is used to check if there should be a pointer
 * modifier appended to a variable.
 * The argument should be a base type.
 */
char *
indamode(type, address_mode)
	struct object *type;
	enum addr address_mode;
{
	if (type->o_class != O_TYPE)
		error(FATAL, "internal error (indamode): not a type");
	switch( address_mode ) {
		case A_VALUE:
			return ("");

		case A_REF:
			return (type->t_constr == C_ARRAY ? "" : "*");
	}
}

/*
 * Names of translation functions for types.
 * Warning: returns pointer to a static buffer.
 */
char *
xfn(kind, type)
	enum translation kind;
	struct object *type;
{
	static char buf[MAXSTR];
	char *name;

	switch (kind) {
	    case EXTERNALIZE:
		name = "externalize";
		break;
	    case INTERNALIZE:
		name = "internalize";
		break;
	}
	(void) sprintf(buf, "%s_%s", name, rep(type));
	return (buf);
}

/*
 * Print the heading for a type externalizing or internalizing function.
 */
xfn_header(kind, type, ptr_type)
	enum translation kind;
	struct object *type, *ptr_type;
{
	FILE *f;

	switch (kind) {
	    case EXTERNALIZE:
		f = support1; break;
	    case INTERNALIZE:
		f = support2; break;
	}
	fprintf(f,
"\n\
static\n\
%s(p, buf)\n\
\tregister %s *p;\n\
\tregister _btype *buf;\n",
		xfn(kind, type), rep(ptr_type));
}

struct object *
enumeration_type(items)
	list items;
{
	struct object *type;
	list p, q;

	type = make_type(C_ENUMERATION);
	type->t_list = items;
	type->t_name = gensym();
	/*
	 * Print a C definition for the enumeration.
	 */
	fprintf(header, "\ntypedef enum {\n");
	for (p = items; p != NIL; p = cdr(p)) {
		q = car(p);
		fprintf(header, "\t%s = %s", ocar(q)->o_name, rep(ocdr(q)));
		if (cdr(p) != NIL)
			fprintf(header, ",\n");
		else
			fprintf(header, "\n");
	}
	fprintf(header, "} %s;\n", type->t_name);
	/*
	 * We use the same sizeof and translation functions
	 * for all enumerated types.
	 */
	type->t_xsize = 4;
	fprintf(header,
"\n\
#define sizeof_%s sizeof_enumeration\n\
#define externalize_%s externalize_enumeration\n\
#define internalize_%s internalize_enumeration\n",
			rep(type), rep(type), rep(type));
	return (type);
}

struct object *
constant_type(const_type, items)
	struct object *const_type;
	list items;
{
	struct object *type;
	list p, q;

	if (const_type->o_class != O_TYPE)
		error(FATAL, "internal error (constant_type): not a type");

	if (const_type->t_constr != C_PREDEF) {
		error(SERIOUS, "type %s is invalid as constant type", rep(const_type));
		return (unspecified_type);
	}

	type = make_type(C_CONSTANT);
	type->t_list = items;
	type->t_name = gensym();
	/*
	 * Print a C definition for the constant type.
	 */
	fprintf(header, "\ntypedef %s %s;\n", rep(const_type), rep(type));
	for (p = items; p != NIL; p = cdr(p)) {
		q = car(p);
		fprintf(header, "#define %s %s\n", ocar(q)->o_name, rep(ocdr(q)));
	}
	/*
	 * The size of the externalized result is taken from the
	 * constant type.
	 */
	type->t_xsize = const_type->t_xsize;
	fprintf(header,
"\n\
#define sizeof_%s sizeof_%s\n\
#define externalize_%s(a,b) externalize_%s(a,b)\n\
#define internalize_%s(a,b) internalize_%s(a,b)\n",
			rep(type), rep(const_type),
			rep(type), rep(const_type),
			rep(type), rep(const_type));
	return (type);
}

struct object *
private_type()
{
	struct object *type;

	type = make_type(C_PRIVATE);
	type->t_list = NIL;
	/* invalid type name so private can only be used in typedef */
	type->t_name = "invalid";
	/* force variable size */
	type->t_xsize = -1;

	return (type);
}

struct object *
record_type(fields)
	list fields;
{
	struct object *type, *t, *bt;
	list p, q;
	int fixed_size;
	char *format, *ref, *member;

	if (fields == NIL) {
		error(SERIOUS, "empty records are not supported");
		return (unspecified_type);
	}
	type = make_type(C_RECORD);
	type->t_list = fields;
	type->t_name = gensym();
	/*
	 * Print a C definition for the record.
	 */
	fprintf(header, "\ntypedef struct {\n");
	for (p = fields; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		for (q = car(car(p)); q != NIL; q = cdr(q))
			fprintf(header, "\t%s %s;\n", rep(t), ocar(q)->o_name); 
	}
	fprintf(header, "} %s;\n", type->t_name);
	/*
	 * Generate sizeof function for the record.
	 * The size is equal to the sum of the sizes of each field.
	 */
	fixed_size = 0;
	type->t_xsize = 0;
	for (p = type->t_list; p != NIL; p = cdr(p)) {
		bt = basetype(ocdr(car(p)));
		if (bt->t_xsize == -1)
			type->t_xsize = -1;
		else
			/* count multiple fields of the same type */
			fixed_size += length(car(car(p))) * bt->t_xsize;
	}
	if (type->t_xsize != -1) {
		/*
		 * The record is fixed-size, so just define a macro.
		 */
		type->t_xsize = fixed_size;
		fprintf(header,
"\n\
#define sizeof_%s(p) %d\n",
			rep(type), type->t_xsize);
	} else {
		/*
		 * There are some variable-size fields, so define a function.
		 */
		fprintf(support1,
"\n\
static int\n\
sizeof_%s(p)\n\
\tregister %s *p;\n\
{\n\
\tregister int size = %d;\n\
\n",
			rep(type), rep(type), fixed_size);
		for (p = type->t_list; p != NIL; p = cdr(p)) {
			t = ocdr(car(p));
			bt = basetype(t);
			if (bt->t_xsize != -1)
				continue;
			ref = refstr(bt);
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(support1,
"\tsize += sizeof_%s(%sp->%s);\n",
					rep(t), ref, ocar(q)->o_name);
		}
		fprintf(support1,
"\treturn (size);\n\
}\n"
			);
	}
	/*
	 * Define translation functions.
	 */
	xfn_header(EXTERNALIZE, type, type);
	xfn_header(INTERNALIZE, type, type);
	format =
"{\n\
\tregister _btype *bp;\n\
\n\
\tbp = buf;\n";
	fprintf(support1, format);
	fprintf(support2, format);
	format =
"\tbp += %s(%sp->%s, bp);\n";
	for (p = type->t_list; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			member = ocar(q)->o_name;
			fprintf(support1, format,
				xfn(EXTERNALIZE, t), ref, member);
			fprintf(support2, format,
				xfn(INTERNALIZE, t), ref, member);
		}
	}
	format =
"\treturn (bp - buf);\n\
}\n";
	fprintf(support1, format);
	fprintf(support2, format);
	return (type);
}

struct object *
error_type(arguments)
	list arguments;
{
	struct object *t;

	t = make_type(C_ERROR);
	t->t_list = arguments;
	return (t);
}

struct object *
procedure_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_PROCEDURE);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

struct object *
execution_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_EXECUTION);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

struct object *
interrupt_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_INTERRUPT);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

struct object *
message_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_MESSAGE);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

struct object *
file_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_FILE);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

struct object *
memory_type(args, results, errors, permission)
	list args, results, errors, permission;
{
	struct object *t;

	t = make_type(C_MEMORY);
	t->t_args = args;
	t->t_results = results;
	t->t_errors = errors;
	t->t_permission = permission;
	return (t);
}

/*
 * Look up the value corresponding to a member of an enumeration type.
 * Print an error message if it's not found.
 */
struct object *
enum_value(symbol, enumtype)
	struct object *symbol, *enumtype;
{
	list p;
	char *name;

	name = symbol->o_name;
	for (p = enumtype->t_list; p != NIL; p = cdr(p))
		if (streq(name, ocar(car(p))->o_name))
			return (ocdr(car(p)));
	error(SERIOUS, "%s not a member of the specified enumeration/constant type", name);
	return (0);
}

/*
 * Return an alias for the candidate.
 * If it is all upper-case, make the alias all lower case.
 * This allows the enumerated constants to be upper case
 * and the union members lower case, conforming to conventional C style.
 */
char *
candidate_name(candidate)
	char *candidate;
{
	register char *cp, *bp;
	int upper_case;
	static char buf[MAXSTR];

	upper_case = 1;
	for (cp = candidate; *cp != '\0'; cp += 1) {
		if (islower(*cp)) {
			upper_case = 0;
			break;
		}
	}
	if (upper_case) {
		strcpy(buf, "u_");
		for (cp = candidate, bp = &buf[2]; *cp != '\0'; cp += 1, bp += 1) {
			if (isupper(*cp))
				*bp = tolower(*cp);
			else
				*bp = *cp;
		}
		*bp = '\0';
	} else {
		sprintf(buf, "%s_variant", candidate);
	}
	return (buf);
}

/*
 * Construct a choice type.
 * There are two ways a choice can be specified:
 * with an explicit enumeration type as a designator,
 * or with an implicit enumeration type,
 * by specifying values as well as names for each designator.
 * Convert the second form into the first by creating
 * an enumeration type on the fly.
 */
struct object *
choice_type(designator, candidates)
	struct object *designator;
	list candidates;
{
	struct object *type, *t;
	list p, q, dlist;
	char *format, *ref, *member;
	int bad = 0;

	if (designator != 0) {
		t = basetype(designator);
		switch( t->t_constr ) {

		case C_ENUMERATION:
		case C_CONSTANT:
			/*
			 * Check that designators don't specify conflicting values.
			 */
			for (p = candidates; p != NIL; p = cdr(p))
				for (q = car(car(p)); q != NIL; q = cdr(q))
					if (cdr(car(q)) != NIL &&
					    true_value(ocdr(car(q))) !=
					    true_value(enum_value(ocar(car(q)), t))) {
						error(SERIOUS, "conflicting value specified for designator %s",
							ocar(car(q))->o_name);
						bad = 1;
						continue;
					}
			break;

		case C_PREDEF:
			/*
			 * Check that designators do specify values.
			 */
			dlist = NIL;
			for (p = candidates; p != NIL; p = cdr(p)) {
				for (q = car(car(p)); q != NIL; q = cdr(q)) {
				       if (cdr(car(q)) == NIL) {
						error(SERIOUS, "value must be specified for designator %s",
							ocar(car(q))->o_name);
						bad = 1;
						continue;
					}
					dlist = nconc(dlist, cons(car(q), NIL));
				}
			}
			if (!bad)
				designator = constant_type(t, dlist);
			break;

		default:
			error(SERIOUS, "designator type %s can not be used in choices",
				rep(designator));
			return (unspecified_type);
		}
	} else {
		/*
		 * Check that designators do specify values.
		 */
		dlist = NIL;
		for (p = candidates; p != NIL; p = cdr(p)) {
			for (q = car(car(p)); q != NIL; q = cdr(q)) {
				if (cdr(car(q)) == NIL) {
					error(SERIOUS, "value must be specified for designator %s",
						ocar(car(q))->o_name);
					bad = 1;
					continue;
				}
				dlist = nconc(dlist, cons(car(q), NIL));
			}
		}
		if (!bad)
			designator = enumeration_type(dlist);
	}
	if (bad)
		return (unspecified_type);
	type = make_type(C_CHOICE);
	type->t_designator = designator;
	type->t_candidates = candidates;
	type->t_name = gensym();
	/*
	 * Print a C definition for the choice.
	 */
	fprintf(header,
"\n\
typedef struct {\n\
\t%s designator;\n\
\tunion {\n",
		rep(designator));
	for (p = candidates; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			member = ocar(car(q))->o_name;
			fprintf(header,
"\t\t%s u_%s;\n\
#define %s u.u_%s\n",
				rep(t), member, candidate_name(member), member);
		}
	}
	fprintf(header,
"\t} u;\n\
} %s;\n",
		type->t_name);
	/*
	 * Generate a sizeof function for the choice.
	 * The size is equal to 1 (for the designator word)
	 * plus the size of the corresponding candidate.
	 * We could check if all the candidates happen to be the same size,
	 * but we don't bother and always call it variable-size.
	 */
	type->t_xsize = -1;
	fprintf(support1,
"\n\
static int\n\
sizeof_%s(p)\n\
\tregister %s *p;\n\
{\n\
\tregister int size = %d;\n\
\n\
\tswitch (p->designator) {\n",
		rep(type), rep(type), designator->t_xsize);
	for (p = candidates; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			member = ocar(car(q))->o_name;
			fprintf(support1,
"\t    case %s:\n\
\t\tsize += sizeof_%s(%sp->%s);\n\
\t\tbreak;\n",
				member, rep(t), ref, candidate_name(member));
		}
	}
	fprintf(support1,
"\t}\n\
\treturn (size);\n\
}\n"
		);
	/*
	 * Define translation functions.
	 */
	xfn_header(EXTERNALIZE, type, type);
	xfn_header(INTERNALIZE, type, type);
	format =
"{\n\
\tregister _btype *bp;\n\
\n\
\tbp = buf + %s(&p->designator, buf);\n\
\tswitch (p->designator) {\n";
	fprintf(support1, format, xfn(EXTERNALIZE, designator));
	fprintf(support2, format, xfn(INTERNALIZE, designator));
	format =
"\t    case %s:\n\
\t\tbp += %s(%sp->%s, bp);\n\
\t\tbreak;\n";
	for (p = candidates; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			member = ocar(car(q))->o_name;
			fprintf(support1, format, member,
				xfn(EXTERNALIZE, t), ref, candidate_name(member));
			fprintf(support2, format, member,
				xfn(INTERNALIZE, t), ref, candidate_name(member));
		}
	}
	format =
"\t}\n\
\treturn (bp - buf);\n\
}\n";
	fprintf(support1, format);
	fprintf(support2, format);
	return (type);
}

/*
 * Find the underlying type of a type.
 */
struct object *
basetype(type)
	struct object *type;
{
	while (type != 0 && type->o_class == O_SYMBOL)
		type = lookup(values, type);
	if (type == 0 || type->o_class != O_TYPE)
		error(FATAL, "internal error (basetype): not a type");
	return (type);
}

/*
 * Make sure a number is a valid constant for this type.
 */
type_check(type, value)
	struct object *type, *value;
{
	struct object *bt;
	enum typeid tid;
	enum constr tconstr;

	if (value->o_class == O_SYMBOL) {
		if (lookup(types, value) != type)
			return (0);
	}
	bt = basetype(type);
	tid = bt->t_id;
	tconstr = bt->t_constr;

	return (tid == T_CARDINAL || /*tid == T_MEDIUM_CARDINAL || */tid == T_LONG_CARDINAL ||
		tid == T_INTEGER || /*tid == T_MEDIUM_INTEGER ||*/ tid == T_LONG_INTEGER ||
		tid == T_UNSPECIFIED || /*tid == T_MEDIUM_UNSPECIFIED ||*/ tid == T_LONG_UNSPECIFIED ||
		tconstr == C_PROCEDURE || tconstr == C_ERROR ||
		tconstr == C_EXECUTION || tconstr == C_FILE ||
		tconstr == C_MESSAGE || tconstr == C_MEMORY);
}


/*
 * Generate client and server functions for procedure declarations.
 */

static int procedure_number = 0;

define_procedure(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_PROCEDURE)
		error(FATAL, "internal error (define_procedure): not a procedure");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	proc_functions(rep(name), t, procedure_number++, random_id);
	declare(&procedures, name, type);
}

/*
 * Generate client and server functions for execution declarations.
 */
define_execution(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_EXECUTION)
		error(FATAL, "internal error (define_execution): not an execution");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	exec_functions(rep(name), t, procedure_number++, random_id);
	declare(&procedures, name, type);
}

/*
 * Generate client and server functions for interrupt declarations.
 */
define_interrupt(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_INTERRUPT)
		error(FATAL, "internal error (define_interrupt): not an interrupt");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	proc_functions(rep(name), t, procedure_number++, random_id);
	declare(&procedures, name, type);
}

/*
 * Generate client and server functions for message declarations.
 */
define_message(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_MESSAGE)
		error(FATAL, "internal error (define_message): not a message");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	message_functions(rep(name), t, procedure_number++, random_id);
	declare(&procedures, name, type);
	name->o_name = "message";
}

/*
 * Generate client and server functions for file declarations.
 */
define_file(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;
	char *oldname;
	char newname[80];

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_FILE)
		error(FATAL, "internal error (define_file): not a file");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	file_functions(rep(name), t, procedure_number, random_id);
	oldname = name->o_name;
	strcpy(newname, oldname );
	strcat(newname, "_write" );
	name->o_name = copy(newname);
	procedure_number++;
	declare(&procedures, name, type);
	strcpy(newname, oldname );
	strcat(newname, "_read" );
	name = make_symbol(newname);
	procedure_number++;
	declare(&procedures, name, type);
}

/*
 * Generate client and server functions for memory declarations.
 */
define_memory(name, type)
	struct object *name, *type;
{
	struct object *t;
	unsigned long random_id;

	if (recursive_flag)
		return;
	t = basetype(type);
	if (t->t_constr != C_MEMORY)
		error(FATAL, "internal error (define_memory): not memory");

	random_id = hash_type( type );
	random_id += (unsigned long)procedure_number;
	type->t_typecheck = random_id;

	memory_functions(rep(name), t, procedure_number++, random_id);
	declare(&procedures, name, type);
	name->o_name = "memory";
}

/*
 * Generate #define's for constant declarations.
 */
define_constant(name, type, value)
	struct object *name, *type, *value;
{
	struct object *t;

	t = basetype(type);
	if (t->t_constr == C_PROCEDURE)
		error(FATAL, "internal error (define_constant): procedure type");
	fprintf(header,
"\n\
#define %s %s\n",
		rep(name), rep(value));
	declare(&values, name, value);
	declare(&types, name, type);
}
