/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 * 18.05.90 kk  disabled turning of/on alarms
 * 29.06.90 kk  removed reset_services() from the default exception handling
 * 21.06.91 kk  renamed member of the exception structure from errno to errorno
 *              because some systems define errno as function
 */

#include "sysconfig.h"
#include <stdlib.h>
#include "ncc.h"
#include "nc.h"


/*
 * Global variables maintained by this module. The queue of active
 * exception handlers, and the corresponding free queue.
 */

struct nc_queue   active_exceptions = { 0 };      /* Active exception handlers */
struct nc_queue   free_exceptions = { 0 };        /* Free exception handlers */
/* kkk disabled turning of/on alarms
extern  int     nc_turnoffalarms;
*/
int     nc_setjmp = 0;                              /* Return from setjmp() */


/*
 * Allocate a descriptor block for an exception handler, and also
 * provide a pointer to the exception buffer for this service. This
 * is used when an exception handler is set up.
 */

int *
nc_allocexception(void)
{
	nct_exhandler *new_exception;

	/* Allocate an exception descriptor for this exception. In this */
	/* call we may get an exception. This is no danger since the    */
	/* preceeding exception handler then will be invoked.           */
	new_exception = (nct_exhandler *)q_rhead( &free_exceptions );
	if( new_exception == 0 ) {
		/* Allocate a new exception handler from the heap.      */
		new_exception = (nct_exhandler *)nc_allocate( sizeof( nct_exhandler ) );
	}

	/* Enqueue this structure on the exception handler queue.       */
	q_atail( &active_exceptions, (struct queue_item *)new_exception );

	/* Return a pointer to the jump buffer within the exception     */
	/* descriptor.                                                  */
	return((int *) &new_exception->nc_jmpbuf[0] );
}


/*
 * Free an exception handler. The one freed is always the one at the
 * tail of the exception queue. It is the users responsibility to free
 * an exception handler when he exits from a procedure.
 */

nct_exhandler *
rfree(void)
{
	nct_exhandler *rm_exception;

	/* Free the exception at the end (tail) of the active queue.    */
	if( (rm_exception = (nct_exhandler *)q_rtail( &active_exceptions )) == 0 ) {
		nc_perror( "rfree(): no exception handler");
		exit(1);
	}

	/* Put the exception handler on the free queue.                 */
	q_atail( &free_exceptions, (struct queue_item *)rm_exception );

	/* Return a pointer to the newly freed exception handler.       */
	return( rm_exception );
}


/*
 * Get a pointer to the exception data structure within an exception
 * handler. This is a structure which is known by the user.
 */

nct_exception *
rexcept(void)
{
	nct_exhandler *rm_exception;

	/* Remove the exception handler from the queue.                 */
	if( (rm_exception = (nct_exhandler *)q_rtail( &active_exceptions )) == 0 ) {
		nc_perror( "rexcept(): no exception handler");
		exit(1);
	}

	/* Put the exception handler back on the active queue.          */
	q_atail( &active_exceptions, (struct queue_item *)rm_exception );

	/* Return a pointer to the actual exception data.               */
	return( &rm_exception->nc_except );
}


/*
 * Raise an exception. In this case the exception handler at the tail
 * of the active queue is invoked by a longjmp(). The error type, code
 * and message is filled in first.
 */

int
nc_raise(int errtype, int errcode, char *errmsg)
{
	nct_exhandler *rm_exception;

	/* Remove the exception handler from the queue.                 */
	if( (rm_exception = (nct_exhandler *)q_rtail( &active_exceptions )) == 0 ) {
		nc_perror( "nc_raise(): no exception handler");
		exit(1);
	}

	/* Update the exception handler with the parameters from the    */
	/* user, and also set errno to the contents of errcode.         */
	rm_exception->nc_except.errtype = errtype;
	rm_exception->nc_except.errorno = errcode;
	rm_exception->nc_except.errmsg = errmsg;
	errno = errcode;

	/* Put the exception handler back on the active queue.          */
	q_atail( &active_exceptions, (struct queue_item *)rm_exception );

	/* Check if the alarms are to be turned off before the jump.    */
/* kkk disabled turning off/on alarms
	if( nc_turnoffalarms ) {
		alarm( 0 );
		nc_turnoffalarms = 0;
	}
*/

	/* Long jump to the exception handler.                          */
	longjmp( rm_exception->nc_jmpbuf, -1 );

	/* If comes here something is really wrong - give fatal error.  */
	nc_perror( "nc_raise(): unable to raise exception");
	exit(1);
}


/*
 * Raise next possible exception. In this case the exception handler at
 * the tail of the active queue is removed, and the preceeding one is
 * invoked. If errtype and errcode both are zero, the same exception
 * is propagated, otherwise the new error is used.
 */

int
rnext(int errtype, int errcode, char *errmsg)
{
	nct_exhandler *rm_exception;

	/* Remove the exception handler from the queue.                 */
	if( (rm_exception = (nct_exhandler *)rfree()) == 0 ) {
		nc_perror( "rnext(): no exception handler");
		exit(1);
	}

	/* If errtype and errcode are zero, the exception is supposed   */
	/* to be propagated. Therefore we have to update the error      */
	/* type, error code and error message.                          */
	if( errtype == 0 && errcode == 0 ) {
		errtype = rm_exception->nc_except.errtype;
		errcode = rm_exception->nc_except.errorno;
		errmsg = rm_exception->nc_except.errmsg;
	}

	/* And finally just raise the exception again.                  */
	nc_raise( errtype, errcode, errmsg );
}
