/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 *
 * xx.05.90 kk  Version 2.0
 * 07.05.90 kk  replaced negative error returns by returning (-1) and setting
 *              errno/errtype
 *              added code related to the rx interface.
 * 04.04.91 Tg  error return modified, round up buffer added
 * 11.09.2000 - CS clean up version
 */

#include "sysconfig.h"
#include "errtype.h"
#include "ncc.h"
#include "nc.h"


/*
 * Send a message of a specified type to the remote server.
 * Returns the call id (the call number in case of rx, or connection in
 * case of connection oriented protocol) if OK, (-1) with errno, errtype
 * set in case of an error
 */
long int
nc_send_call(int mtype, int service_entry_id, unsigned int proc_number, long unsigned int random_id, _btype *buf, int buffer_size)
{
	register nct_cservice *cd;
	ncct_call_header *ncc_call;
	ncct_return_header ncc_return;
	int nbytes, status;
	protocol *p;

	/* Get a pointer to the connection descriptor.                  */
	if( (service_entry_id < 0) || ((cd = nc_sgetptr( service_entry_id )) == 0) ) {
		nc_deallocate(buf);
		RETURNERR( ENCNOSERVICE, ETNCRT, "nc_sgetptr()" );
	}


/* kkk in future we may preset header in cd and copy it */
	/* Set the protocol header.                             */
	ncc_call = (ncct_call_header *)buf;
	ncc_call->ncc_byteorder         = NC_BYTEORDER;
	ncc_call->ncc_floattype         = NC_FLOATTYPE;
	ncc_call->ncc_version           = nc_version;
	ncc_call->ncc_revision          = nc_revision;
	ncc_call->ncc_length            = buffer_size;
	ncc_call->ncc_random_id         = random_id;
	ncc_call->ncc_service_id        = cd->ncc_service;
	ncc_call->ncc_procedure_number  = proc_number;
	ncc_call->ncc_userid            = cd->ncc_uid;
	ncc_call->ncc_groupid           = cd->ncc_gid;
	ncc_call->ncc_type              = mtype;
	ncc_call->ncc_errno             = errno;

	TRACE(callmsg && fprintf(stderr, "nc_send_call type %d to %s on sd %d: mt %d %d v%d.%d len %d s_id %d pr %d err %d\n",
		mtype, print_ipaddr(&(cd->ncc_cd.host)), service_entry_id,
		ncc_call->ncc_byteorder, ncc_call->ncc_floattype, ncc_call->ncc_version,
		ncc_call->ncc_revision, ncc_call->ncc_length - CALL_HEADER_BYTES , ncc_call->ncc_service_id,
		ncc_call->ncc_procedure_number,ncc_call->ncc_errno ) );

	/* TRACE data */
	if (nctrace.tron && nctrace.data) {
		char *p;
		int i;

		fprintf(stderr, "nc_call(): call data:");
		for (i = 0, p = buf+CALL_HEADER_BYTES; i<buffer_size-CALL_HEADER_BYTES; ++i, ++p) {
			if (!(i%20)) fprintf(stderr, "\n");
			fprintf(stderr, "%2x ",(*p & 0xff));
		}
		fprintf(stderr, "\n");
	}
	/* Update the statistics variable */
	nc_statistics.cl_calls++;
	nc_statistics.cl_sendlen += (buffer_size - CALL_HEADER_BYTES);

	/* Send the buffer to the remote side. Do not wait for the      */
	/* response to come back.                                       */
	p = cd->ncc_protocol;
	if (p == NC_P_RX) {
		/* we are using this call only for non-blocking RPC and messages */
		cd->ncc_cd.smsg = buf;
		cd->ncc_cd.rmsg = 0;
		cd->ncc_cd.slen = buffer_size;
		cd->ncc_cd.rlen = 0;
		cd->ncc_cd.flags = cd->ncc_callflags;
		if (mtype == NCC_NOREPLY)
			cd->ncc_cd.flags |= RX_NOREPL;
		cd->ncc_cd.timeout = cd->ncc_noresponse;
		if (mtype == NCC_UNRELIABLE)
			status = rx_unrel(&(cd->ncc_cd));
		else
			status = rx_call(&(cd->ncc_cd));
		nc_deallocate(buf);
		if (status < 0)
			return(-1);
		else
			return(cd->ncc_cd.call_number);
	} else {

                return(-1);
	}
}


/*
 * Receive a message of a certain type from the remote server.
 * Returns buffer with data if OK, NULL with errno, errtype set in case of an error
 */

_btype *
nc_get_reply(int mtype, int service_entry_id, _btype *inbuf, int inlen)
{
	register nct_cservice *cd;
	register _btype *rp;
	ncct_return_header ncc_return;
	int length;
	protocol *p;
	int nbytes, status;
	long timeout;

	/* Get a pointer to the connection descriptor.                  */
	if( (service_entry_id < 0) || ((cd = nc_sgetptr( service_entry_id )) == 0) ) {
		RETURNERRP((_btype*)0, ENCNOSERVICE, ETNCRT, "nc_sgetptr()" );
	}

	/* Receive the return header to get the total length of         */
	/* the user data and other parameters.                          */
	if ((p = cd->ncc_protocol) == NC_P_RX) {
		cd->ncc_cd.rmsg = (char *)&ncc_return;
		cd->ncc_cd.rlen = RETURN_HEADER_BYTES;
		if ((timeout = cd->ncc_noresponse) == 0)
			timeout = 60000;        /* one minute */
		if ((status = rx_wait(&(cd->ncc_cd.call_number), timeout)) < 0)
			return(0);
		if (status == 0)
			RETURNERRP((_btype*)0, ENCCNOREPL, ETNCRT, "rx_wait()" );
		if ((status = rx_get_reply(&(cd->ncc_cd))) < 0)
			return(0);
		if (status == 0)
			RETURNERRP((_btype*)0, ENCCNOREPL, ETNCRT, "rx_get_reply()" );
	} else {
                return((_btype *)(-1));
	}
	/* Internalize the return header for later use.                 */
	if (ncc_return.ncc_byteorder != NC_BYTEORDER) {
		ncc_return.ncc_length = swapal( ncc_return.ncc_length );
		ncc_return.ncc_errtype = swapas( ncc_return.ncc_errtype );
		ncc_return.ncc_errcode = swapas( ncc_return.ncc_errcode );
		ncc_return.ncc_type = swapas( ncc_return.ncc_type );
	}
	errno = ncc_return.ncc_errcode;
	TRACE(retmsg && fprintf(stderr, "nc_get_reply from %s on sd %d: mt %d %d v%d.%d len %d err_t %d err %d\n",
		print_ipaddr(&(cd->ncc_cd.host)), service_entry_id,
		ncc_return.ncc_byteorder, ncc_return.ncc_floattype, ncc_return.ncc_version,
		ncc_return.ncc_revision, ncc_return.ncc_length  - RETURN_HEADER_BYTES, ncc_return.ncc_errtype,
		ncc_return.ncc_errcode) );


	if ((rp = nc_allocate( length = ncc_return.ncc_length - RETURN_HEADER_BYTES )) == (char *)0)
		return((char *) 0);

	/* Update the statistics variable */
	nc_statistics.cl_receivelen += length;

	/* The allocation was OK, so we can read in the rest of */
	/* the outstanding data.                                */
	/*if ((p = cd->ncc_protocol) == NC_P_RX)*/
	/* CS 11.09.2000 procotl is RX only */
	 {
		cd->ncc_cd.rlen = length;
		cd->ncc_cd.rmsg = rp;
		if(length > 0 && rx_more(&(cd->ncc_cd)) < 0 ) {
			nc_deallocate(rp);
			return((char *)0);
		}
                /* TRACE data */
                if (nctrace.tron && nctrace.data) {
                        char *p;
                        int i;

                        fprintf(stderr, "nc_get_reply(): return data:");
                        for (i = 0, p = rp; i<cd->ncc_cd.rlen; ++i, ++p) {
                                if (!(i%20)) fprintf(stderr, "\n");
                                fprintf(stderr, "%2x ",*p);
                        }
                fprintf(stderr, "\n");
                }

	}
	/* CS 11.09.2000 branch never taken!!
	 else {

	}
        */
	/* Check the status to see if everything is OK.                 */
	if( ncc_return.ncc_errtype != 0 ) {
		nc_deallocate(ncerrstring); /* keep only last err string */
		internalize_string( &ncerrstring, rp );
		nc_deallocate( rp );
		RETURNERRP((_btype*)0, ncc_return.ncc_errcode,  ncc_return.ncc_errtype, ncerrstring )
	}

	/* Return the buffer to the calling stub.                       */
	return( rp );
}


/*
 * Waits for asynch replies until timeout. If call id is 0 it will return
 * id of the first call found in call_id. The call id is the same as returned
 * by nc_send_call(). Then it's users responsability to
 * find the match with the service descriptor.
 * If p is equal to NC_P_FIRST, the protocol of the first exported service
 * will be used.
 */

int
nc_wait_reply(long int call_id, protocol *p, long int timeout)
{
	register nct_cservice *cd;
	int sid, status;
	unsigned long call_number;

	if (p == NC_P_FIRST) {
		/* Check if any services are imported at all.           */
		if( (sid = nc_sfirst( NC_IMPORTED )) < 0 ) {
			return(-1);
		}
		/* Get the current protocol to be used. For now just            */
		/* use the protocol given in the first service.                 */
		if( (cd = nc_sgetptr( sid )) == 0 ) {
			RETURNERR( ENCNOSERVICE, ETNCRT, "nc_sgetptr()" )
		}
		if ((p = cd->ncc_protocol) == 0) {
			RETURNERR(ENCCNOPROT, ETNCRT, "")
		}
	}

	if (p == NC_P_RX) {
		call_number = call_id;
		if ((status = rx_wait(&call_number, timeout)) < 0)
			return(-1);
		if (status == 0)
			return(0);
		return(call_number);
	} else
		/* should put the normal select here */
		RETURNERR( 0, ETNCRT, "not yet implemented for this protocol")
}

int 
nc_abort_call(int service_entry_id)
{
	nct_cservice *cd;
        protocol *p;

        /* Get a pointer to the connection descriptor.                  */
        if( (service_entry_id < 0) || ((cd = nc_sgetptr( service_entry_id )) ==
0) ) {
                RETURNERR( ENCNOSERVICE, ETNCRT, "nc_sgetptr()" );
        }
        p = cd->ncc_protocol;
        if (p == NC_P_RX) {
		rx_abort(&(cd->ncc_cd));
	} 
}
